/*
 * debug.c - Debugger interface
 *
 * Written 2003 by Werner Almesberger, Caltech Netlab FAST project
 */

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <assert.h>

#include "dwarfx.h"
#include "ptrace.h"

#include "util.h"
#include "error.h"
#include "data.h"
#include "var.h"
#include "op.h"
#include "stmt.h"
#include "sbug.h"


extern void open_elf(const char *name);

extern struct dwx_obj *objects;


static struct process {
    struct pt_process *proc;
    struct breakpoint *breakpoints;
    struct object *objects;
    struct process *next;
} *subprocesses = NULL;

struct breakpoint {
    struct process *process; /* does NOT count as a reference;
				NULL if process is dead */
    unsigned long addr;
    struct breakpoint *next; /* only valid if process != NULL */
};

struct object {
    struct process *process;
    const struct dwx_obj *obj;
    unsigned long addr;
    struct object *next;
};


struct dwx_obj *objects = NULL;


/* ----- Helper functions -------------------------------------------------- */


static int process_has_terminated(const struct process *process)
{
    return process->proc->terminated;
}


static void check_process_termination(const struct process *process)
{
    if (process_has_terminated(process)) clerrorf("process has terminated");
}


/* ----- Values ------------------------------------------------------------ */


struct dwx_type process_type;
struct dwx_type breakpoint_type;
struct dwx_type object_type;


struct value *process_value(struct process *process)
{
    struct value *value;

    value = new_value(&process_type);
    value->value = process;
    return value;
}


struct process *__value_process(struct value *value)
{
    if (!value || value->type != &process_type)
        clerrorf("invalid type (expected process)");
    return (struct process *) value->value;
}


struct value *breakpoint_value(struct breakpoint *breakpoint)
{
    struct value *value;

    value = new_value(&breakpoint_type);
    value->value = breakpoint;
    return value;
}


struct breakpoint *__value_breakpoint(struct value *value)
{
    if (!value || value->type != &breakpoint_type)
        clerrorf("invalid type (expected breakpoint)");
    return (struct breakpoint *) value->value;
}


struct value *object_value(struct object *obj)
{
    struct value *value;

    value = new_value(&object_type);
    value->value = obj;
    return value;
}


const struct object *__value_object(struct value *value)
{
    if (!value || value->type != &object_type)
        clerrorf("invalid type (expected object)");
    return value->value;
}


int sbug_put(struct value *value)
{
    if (value->type == &process_type) {
	struct process *process = (struct process *) value->value;
	struct process **walk;
	struct breakpoint *bp;
	struct object *object;

	pt_terminate(process->proc);
	for (walk = &subprocesses; *walk != process; walk = &(*walk)->next);
	*walk = process->next;
	for (bp = process->breakpoints; bp; bp = bp->next) {
	    assert(bp->process == process);
	    bp->process = NULL;
	}
	for (object = process->objects; object; object = object->next) {
	    assert(object->process == process);
	    object->process = NULL;
	}
	free(process);
	return 1;
    }
    if (value->type == &breakpoint_type) {
	struct breakpoint *bp = (struct breakpoint *) value->value;

	if (bp->process) {
	    struct breakpoint **walk;

	    if (!process_has_terminated(bp->process))
		pt_unbreak(bp->process->proc,bp->addr);
	    for (walk = &bp->process->breakpoints; *walk != bp;
	      walk = &(*walk)->next);
	    *walk = bp->next;
	}
	free(bp);
	return 1;
    }
    if (value->type == &object_type) {
	struct object *object = (struct object *) value->value;

	if (object->process) {
	    struct object **walk;

	    for (walk = &object->process->objects; *walk != object;
	      walk = &(*walk)->next);
	    *walk = object->next;
	}
	free(object);
	return 1;
    }
    return 0;
}


/* ----- Operators: run ---------------------------------------------------- */


static int need_atexit = 1;


static void kill_subprocesses(void)
{
    struct process *process;

    for (process = subprocesses; process; process = process->next)
	pt_terminate(process->proc);
}


static char *lookup_program(char *name)
{
    const char *path,*pos,*colon;
    int len;

    if (strchr(name,'/')) return stralloc(name);
    path = getenv("PATH");
    if (!path) goto not_found;
    len = strlen(name);
    for (pos = path; *pos; pos = colon+1) {
	const char *end;
	char *tmp;

	colon = strchr(pos,':');
	if (colon) end = colon;
	else {
	    end = strchr(pos,0);
	    colon = end-1;
	}
	tmp = alloc(end-pos+len+2);
	    /* +2: / and \0 */
	memcpy(tmp,pos,end-pos);
	tmp[end-pos] = '/';
	strcpy(tmp+(end-pos)+1,name);
	if (!access(tmp,R_OK | X_OK)) return tmp;
	free(tmp);
    }
not_found:
    clerrorf("%s: command not found",name);
}


static void load_symbols(const char *path)
{
    struct dwx_obj **link = &objects;

    open_elf(path);
    while (1) {
	const void *p;

	p = dwarf_comp_unit();
	if (!p) break;
	*link = dwx_read_objects(p);
	link = &(*link)->next;
    }
}


static struct process *launch_process(const char *path,char **argv)
{
    struct process *process;

    process = alloc_t(struct process);
	/* allocate before fork, so we're sure it works */
    process->proc = pt_fork();
    if (process->proc) {
	process->breakpoints = NULL;
	process->objects = NULL;
	process->next = subprocesses;
	subprocesses = process;
	if (need_atexit) {
	    atexit(kill_subprocesses);
	    need_atexit = 0;
	}
	return process;
    }
    else {
	if (execvp(path,argv) < 0) {
	    perror("execvp");
	    exit(1);
	}
	/* can't happen */
	exit(0);
    }
}


struct value *op_run(const struct op *self)
{
    struct arg_list *arg;
    int args;
    struct value *value0,*value;
    char **argv;
    char *path;

    args = 0;
    for (arg = self->u.run_args.args; arg; arg = arg->next)
	args++;
    argv = alloc((args+2)*sizeof(const char *));
	/* +2: argv[0], terminating NULL */
    value0 = evaluate(self->u.run_args.cmd);
    argv[0] = (char *) __value_string(value0);
    args = 1;
    for (arg = self->u.run_args.args; arg; arg = arg->next) {
	assert(!arg->value);
	arg->value = evaluate(arg->expr);
	argv[args] = (char *) __value_string(arg->value);
	args++;
    }
    argv[args] = NULL;
    path = lookup_program(*argv);
    load_symbols(path);
    value = process_value(launch_process(path,argv));
    free(path);
    for (arg = self->u.run_args.args; arg; arg = arg->next) {
	put(arg->value);
	arg->value = NULL;
    }
    put(value0);
    free(argv);
    return value;
}


/* ----- Operators: break  ------------------------------------------------- */


static struct value *name_to_object(const struct dwx_obj *base,
  const char *name)
{
    const struct dwx_obj *obj;
    struct object *object;

    obj = dwx_find_item(base,name);
    if (!obj) clerrorf("no such object \"%s\"",name);
    if (!subprocesses) clerrorf("no subprocess");
    object = alloc_t(struct object);
    object->process = subprocesses;
    object->obj = obj;
    object->addr = obj->low_pc;
    object->next = object->process->objects;
    object->process->objects = object;
    return object_value(object);
}


static unsigned long addr_eval(const struct op *op)
{
    struct value *value;
    const struct object *object;
    unsigned long addr;

    value = evaluate_lvalue(op);
    if (!value) clerrorf("undefined address");
    if (value->type == &string_type)
	value = name_to_object(objects->children,value_string(value));
    if (value->type != &object_type) return value_int(value);
    object = __value_object(value);
    if (object->obj->kind != dwx_obj_fn && object->obj->kind != dwx_obj_label)
	clerrorf("invalid object type (expected function or label)");
    addr = object->obj->low_pc;
    put(value);
    return addr;
}


struct value *op_breakpoint(const struct op *self)
{
    struct value *value;
    struct process *process;
    unsigned long addr;
    struct breakpoint *bp;

    if (self->u.arg.a) {
	value = evaluate(self->u.arg.a);
	process = __value_process(value);
    }
    else {
	if (!subprocesses) clerrorf("no subprocess");
	process = subprocesses;
    }
    check_process_termination(process);
    addr = addr_eval(self->u.arg.b);
    pt_break(process->proc,addr);
    bp = alloc_t(struct breakpoint);
    bp->process = process;
    bp->addr = addr;
    bp->next = process->breakpoints;
    process->breakpoints = bp;
    if (self->u.arg.a) put(value);
    return breakpoint_value(bp);
}


/* ----- Operators: object access ------------------------------------------ */


/*
 * Design issue: is this the right moment for binding an object to an address
 * space (process) ? Or should this be deferred to the time of evaluation ?
 * Normally, there isn't much of a difference, of course.
 */


struct value *op_object(const struct op *self)
{
    const char *path = self->u.arg.a ? string_eval(self->u.arg.a) : NULL;
    const char *name = self->u.arg.b ? string_eval(self->u.arg.b) : NULL;
    const char *object_name = string_eval(self->u.arg.c);
    const struct dwx_obj *obj;

    obj = objects;
    if (path) {
	obj = dwx_find_unit(objects,path);
	if (!obj) clerrorf("no such unit \"%s\"",path);
    }
    if (name) {
	obj = dwx_find_item(obj->children,name);
	if (!obj) clerrorf("no such parent object \"%s\"",name);
    }
    return name_to_object(obj->children,object_name);
}


/* ----- Operators: destructor --------------------------------------------- */


int destroy_sbug_op(const struct op *op)
{
    if (op->fn == op_run) {
	destroy_expr(op->u.run_args.cmd);
	destroy_arg_list(op->u.run_args.args);
	return 1;
    }
    return 0;
}


/* ----- Statements -------------------------------------------------------- */


void stmt_continue_process(const struct stmt *self)
{
    struct value *value;
    struct process *process;

    if (self->u.process.process) {
	value = evaluate(self->u.process.process);
	process = __value_process(value);
    }
    else {
	if (!subprocesses) clerrorf("no subprocess");
	process = subprocesses;
    }
    check_process_termination(process);
    pt_run(process->proc);
    if (self->u.process.process) put(value);
    /*
     * @@@ Not very flexible ...
     */
    (void) pt_wait(1);
}


int destroy_sbug_stmt(const struct stmt *stmt)
{
    if (stmt->fn == stmt_continue_process) {
	destroy_expr(stmt->u.process.process);
	return 1;
    }
    return 0;
}


/* ----- Data access ------------------------------------------------------- */


struct value *dereference_object(struct value *value)
{
    const struct object *object;
    const struct dwx_type *type;
    int size;
    struct value *new;
    void *data;

    object = __value_object(value);
    if (!object->process) clerrorf("referenced process no longer exists");
    if (object->obj->kind != dwx_obj_var)
	clerrorf("invalid object type (expected variable or constant)");
    type = object->obj->u.var.type;
    size = dwx_sizeof(type);
    data = alloc(size);
    /*
     * @@@ Optimize: create global owner of constant data, then use this as
     * new->owner
     */
    if (object->obj->u.var.constant)
	memcpy(data,object->obj->u.var.constant,size);
    else {
	check_process_termination(object->process);
	pt_read(object->process->proc,
	  dwx_loc_here(object->obj->u.var.loc,object->process->proc->regs.eip),
	  data,size);
		/* ia32-ism @@@ */
    }
    put(value);
    new = new_value(type);
    new->value = data;
    return new;
}
